/**
  \file  properties.h
  \brief Load properties from config file
  
  the config file is searched in the current directory and loads
  the defaults for creating the nouveau_reg.h file.
  
*/
#ifndef _PROPERTIES_H
#define _PROPERTIES_H

#define PROPERTIES_FILENAME  ".generaterc"		///< Filename to store properties in

/** \defgroup pr_com Comment property

 Which comments are inserted:
   COMMENTS_NONE:     Only the comment "used for NV04 NV30 NV40 ..." is printed
   COMMENTS_SHORT:    NONE + parameter info for the define on the same line
   COMMENTS_VERBOSE:  Adds all possible comments (WARNING: HUGE!)
 Please note, that comments configured in the ressource file will show up
 with both VERBOSE and SHORT
   
\{
*/
#define COMMENTS_NONE   	0 		
#define COMMENTS_SHORT  	1
#define COMMENTS_VERBOSE	2
/** \} */
/** \typedef t_entry
    \brief   Holds one entry for the variable config
    
    You can define some define names to get some more infos in the comments
    So basically you'll get:
    #define en_name   value / * en_strval. Added infos from nv_objects... * /
*/
typedef struct entry {
                       char          en_name[100];    ///< Name of the define
                       struct entry *en_next;         ///< Simple list
                       char          en_strval[250];  ///< Value to add to the comment
                     } t_entry;

typedef struct {char   pr_author[50];		///< EMailadresse / Name of the creator
                int    pr_version;		///< Print objects.c / nv_objects.h version true/false
                int    pr_time;			///< Boolean: Print time of creation
                int    pr_comments;		///< What type of comments?
                char   pr_licensefile[200];	///< Filename of the license text to add "" if not used
                char   pr_include[200];         ///< Include file for constant defines, comments etc. 
                                                ///< Will be copied verbatim from file into nouveau_drm.h
                t_entry *pr_list;		///< List contains additional name / comments pairs
               } t_properties;

/**
 \fn	t_properties *getProperties(int argc, char *argv[]);
 \brief	reads the rc file or defines defaults
 \return Pointer to the properties read (or defaults)
*/
t_properties *getProperties(int argc, char *argv[]);

/**
 \fn 	t_properties *readRcFile(void);
 \brief read properties from PROPERTIES_FILENAME
 \return either the properties read or NULL if there was no such file
*/
t_properties *readRcFile(void);

/**
 \fn	t_properties *getDefaultProperties(void);
 \brief define a structure with default values
 \return default properties
*/
t_properties *getDefaultProperties(void);

/** 
 \fn 	void get_version_line(char *version)
 \brief Copy the version of objects.c into version 
 \param version needs to hold at least 21 bytes
*/
void get_version_line(char *version);
#endif

