/*
	Renouveau shader parser
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <stdint.h>

#include "rexml.h"

const char *filename_dump = NULL;
const char *filename_dbxml = "renouveau.xml";
const char *shader_type = NULL;

void print_help();
void parse_dump(const char *filename);

void rexmlParseBitfields(xmlNodePtr node, int value, char *parsed_value);
void rexmlFillEnum(xmlNodePtr node, int value, char *parsed_value);

int main(int argc, char **argv)
{
	int i;

        if (argc == 1)
        {
                print_help();
                exit(0);
        }

	for(i=1; i<argc; i++) {
		if (strcmp("-dump", argv[i])==0 || strcmp("--dump", argv[i])==0) {
			filename_dump = argv[++i];
                        continue;
		}
                else if (strcmp("-dbxml", argv[i])==0 || strcmp("--dbxml", argv[i])==0) {
			filename_dbxml = argv[++i];
                        continue;
		}
                else if (strcmp("-type", argv[i])==0 || strcmp("--type", argv[i])==0) {
			shader_type = argv[++i];
                        continue;
		}
                else if (strcmp("-help", argv[i])==0 || strcmp("--help", argv[i])==0) {
                        print_help();
			exit(0);
		}
                else {
                        print_help();
                        exit(0);
                }
	}

	if (rexmlInit(filename_dbxml)<0) {
		fprintf(stderr, "Can not load xml database %s\n", filename_dbxml);
		return 1;
	}

	if (filename_dump) {
		parse_dump(filename_dump);
	}

	rexmlShutdown();
	return 0;
}

void print_help()
{
        printf("Requires:\n");
        printf(" card_10de-xxxx_shader_programs.txt\n");
        printf("Parameters (accepts both - and --):\n");
        printf(" -dbxml <filename.xml> (xml database for parser)\n");
        printf(" -dump <filename-dump.txt> (dump to parse)\n");
        printf(" -type <shader_type> (NVX0_FP for frag progs, NVX0_VP for vtx progs)\n");
}

void parse_dump(const char *filename)
{
	FILE *f;
	char tmp_str[512];
	int instruction[4];
	xmlXPathObjectPtr renouveau_path = NULL;
	xmlNodePtr prog_inst_node = NULL, prog_nodes[4]={NULL, NULL, NULL, NULL};
	xmlNodeSetPtr nodeset;

	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return;
	}

	renouveau_path = rexmlSearchNode(BAD_CAST "/renouveau");
	if (!renouveau_path) {
		fprintf(stderr, "Can not find xml path\n");
		fclose(f);
		return;
	}

	nodeset = renouveau_path->nodesetval;
	if (!xmlXPathNodeSetIsEmpty(nodeset)) {
		prog_inst_node = rexmlSearchNodeAttrib(nodeset->nodeTab[0], "program", "name", shader_type);
	}

	prog_nodes[0] = rexmlSearchNodeAttrib(prog_inst_node, "reg32", "offset", "0x0000");
	prog_nodes[1] = rexmlSearchNodeAttrib(prog_inst_node, "reg32", "offset", "0x0004");
	prog_nodes[2] = rexmlSearchNodeAttrib(prog_inst_node, "reg32", "offset", "0x0008");
	prog_nodes[3] = rexmlSearchNodeAttrib(prog_inst_node, "reg32", "offset", "0x000c");

	if (!prog_inst_node || !prog_nodes[0] || !prog_nodes[1] || !prog_nodes[2] || !prog_nodes[3]) {
		fprintf(stderr, "Can not find node for fifo instruction\n");
		xmlXPathFreeObject(renouveau_path);
		fclose(f);
		return;
	}

	while (!feof(f)) {
		/* Skip comments */
		if (fgetc(f)=='#') {
			int c = '#';
			unsigned int pos = 0;

			memset(tmp_str, 0, sizeof(tmp_str));
			while (c != '\n') {
				if (pos<sizeof(tmp_str)-1) {
					tmp_str[pos++] = c;
				}
				c = fgetc(f);
			}

			printf("%s\n", tmp_str);
		}

		if (fscanf(f, "%08x %08x %08x %08x", &instruction[0], &instruction[1], &instruction[2],
		    &instruction[3]) != 4)
		{
			continue;
		}

		printf("%08x %08x %08x %08x\t", instruction[0], instruction[1], instruction[2],
			instruction[3]);

		memset(tmp_str, 0, sizeof(tmp_str));
		rexmlParseBitfields(prog_nodes[0], instruction[0], tmp_str);
		printf("  %s", tmp_str);
		memset(tmp_str, 0, sizeof(tmp_str));
		rexmlParseBitfields(prog_nodes[1], instruction[1], tmp_str);
		printf(" %s", tmp_str);
		memset(tmp_str, 0, sizeof(tmp_str));
		rexmlParseBitfields(prog_nodes[2], instruction[2], tmp_str);
		printf(" %s", tmp_str);
		memset(tmp_str, 0, sizeof(tmp_str));
		rexmlParseBitfields(prog_nodes[3], instruction[3], tmp_str);
		printf(" %s\n", tmp_str);
	}

	fclose(f);

	xmlXPathFreeObject(renouveau_path);
}

/*
	Parse bitfields of a node
*/

void rexmlParseBitfields(xmlNodePtr node, int value, char *parsed_value)
{
	xmlNodePtr arrsz = node->children;
	xmlChar *attValue, *fieldname, *fieldtype;
	char tmp[32];
	int highbit = -1, lowbit = -1, field, mask, firsttime = 1;
	uint32_t bitfield_mask_used = 0;

 	for(; arrsz; arrsz=arrsz->next) {
		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "bitfield")!=0) {
			continue;
		}

		fieldname = xmlGetProp(arrsz, BAD_CAST "name");
		if (!fieldname) {
			continue;
		}

		fieldtype = xmlGetProp(arrsz, BAD_CAST "type");

		attValue = xmlGetProp(arrsz, BAD_CAST "high");
		if (attValue) {
			highbit = atoi((char*)attValue);
			xmlFree(attValue);
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "low");
		if (attValue) {
			lowbit = atoi((char*)attValue);
			xmlFree(attValue);
		}

		if ((lowbit<0) || (highbit<0)) {
			continue;
		}

		mask = (highbit==31) ? -1 : (1<<(highbit+1))-1;
		field = ((unsigned int) (value & mask)) >> lowbit;
		bitfield_mask_used |= ((unsigned int) mask) & (0xffffffffUL << lowbit);

		/* Add separator if needed */
		if (!firsttime) {
			strcat(parsed_value, " | ");
		}

		strcat(parsed_value, fieldname);
		strcat(parsed_value, "=");

		/* Default to hexa */
		sprintf(tmp, "0x%08x", field);
		if (xmlStrcmp(fieldtype, BAD_CAST "int")==0) {
			sprintf(tmp, "%d", field);
		} else if (xmlStrcmp(fieldtype, BAD_CAST "boolean")==0) {
			sprintf(tmp, "%s", field ? "TRUE" : "FALSE");
		} else if (xmlStrcmp(fieldtype, BAD_CAST "program")==0) {
			sprintf(tmp, "0x%08x", field<<lowbit);
		} else if (xmlStrcmp(fieldtype, BAD_CAST "enum")==0) {
			rexmlFillEnum(arrsz, field, tmp);
		}
		strcat(parsed_value, tmp);

		if (fieldtype) {
			xmlFree(fieldtype);
		}
		xmlFree(fieldname);
		firsttime = 0;
	}
}

/*
	Fill value from enumeration
*/

void rexmlFillEnum(xmlNodePtr node, int value, char *parsed_value)
{
	xmlXPathObjectPtr pathobj;
	xmlNodePtr list, arrsz;
	xmlChar *attValue;

	xmlChar *enum_name = xmlGetProp(node, BAD_CAST "enum_name");
	if (!enum_name) {
		return;
	}

	pathobj = rexmlSearchNode(BAD_CAST "/renouveau/enums");
	if (!pathobj) {
		xmlFree(enum_name);
		return;
	}

	/* Search an enumeration list with this name */
	list = rexmlSearchNodeAttrib(pathobj->nodesetval->nodeTab[0],
		"list", "name", (char*)enum_name);

	if (!list) {
		xmlXPathFreeObject(pathobj);
		xmlFree(enum_name);
		return;
	}

	/* Now search an item with this value */
	arrsz = list->children;
 	for(; arrsz; arrsz=arrsz->next) {
		int found = 0, item_value = 0;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "item")!=0) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "value");
		if (attValue) {
			if (sscanf((char*)attValue, "0x%x", &item_value)) {
				found = (item_value == value);
			}
			xmlFree(attValue);
		}

		if (found) {
			attValue = xmlGetProp(arrsz, BAD_CAST "name");
			if (attValue) {
				sprintf(parsed_value, "%s", attValue);
				xmlFree(attValue);
			}
			break;
		}
	}

	xmlXPathFreeObject(pathobj);
	xmlFree(enum_name);
}
