/*
	Renouveau dump parser
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rexml.h"
#include "nv_hw.h"

typedef struct {
	int name;
	int context;
	int data[4];
	int engine;
} obj_info_t;

#define SUBCHAN_NAME_SZ 128

typedef struct {
	int hash;
	xmlNodePtr node;
	char name[SUBCHAN_NAME_SZ];
} subchan_t;

typedef struct {
	int value;
	const char *name;
} glenum_t;

obj_info_t *obj_infos=NULL;
int num_obj=0;

int pci_id = 0;
int boot0 = 0;
int card_family;
int obj_from_hash = 0;
int print_leftover = 0;
int print_as_floats = 0;
int got_objects = 0;
subchan_t subchan[8];
char file_startup[512];
int dump_value_float;
int dump_simple_value;
char parsed_value[512]; /* Where we store parsed dump value */

void get_pci_id(const char *filename);
void get_boot0(const char *filename);
int read_objects(const char *filename);
void add_object(int name, int context, int *data);
void parse_dump(const char *filename, int readsubchan);
int get_object_class(int channel);
int get_object_class_from_hash(int hash);

int main(int argc, char **argv)
{
	const char *filename_dump = NULL;
	const char *filename_objects = "card_objects.txt";
	const char *filename_stdout = "card_stdout.txt";
	const char *filename_startup = "card_%04x-%04x_test_startup.txt";
	const char *filename_dbxml = "renouveau.xml";
	int i;

	for(i=0; i<argc; i++) {
		if (strcmp("-dump", argv[i])==0) {
			filename_dump = argv[i+1];
		}
		if (strcmp("-dbxml", argv[i])==0) {
			filename_dbxml = argv[i+1];
		}
		if (strcmp("-objfromhash", argv[i])==0) {
			obj_from_hash = 1;
		}
		if (strcmp("-pciid", argv[i])==0) {
			sscanf(argv[i+1], "%x", &pci_id);
		}
		if (strcmp("-leftover", argv[i])==0) {
			print_leftover = 1;
		}
		if (strcmp("-floats", argv[i])==0) {
			print_as_floats = 1;
		}
		/*if (strcmp("-objects", argv[i])==0) {
			filename_objects = argv[i+1];
		}
		if (strcmp("-stdout", argv[i])==0) {
			filename_stdout = argv[i+1];
		}*/
		if (strcmp("-help", argv[i])==0) {
			printf("Requires:\n");
			printf(" card_objects.txt, card_stdout.txt\n");
			printf("Optional:\n");
			printf(" card-PCIID-test_startup.txt for your card\n");
			printf(" in current directory (if objfromhash not used)\n");
			printf("Parameters:\n");
			printf(" -dbxml <filename.xml> (xml database for parser)\n");
			printf(" -dump <filename-dump.txt> (dump to parse)\n");
			printf(" -objfromhash (use object type from hash value)\n");
			/* hashobjtype = XX in a yyyyXXyy hash */
			printf(" -pciid <n> (force pci id, must be 10deXXXX)\n");
			printf(" -subchan <n> <hashobjtype> (force hash, use with objfromhash)\n");
			printf(" -leftover (print remaining unknown bits)\n");
			printf(" -floats (print values as floats)\n");
			exit(0);
		}
	}

	/* Retrieve card family */
	if (pci_id == 0) {
		get_pci_id(filename_stdout);
	}
	get_boot0(filename_stdout);
	card_family = get_card_family(boot0);
	if (card_family < 0) {
		fprintf(stderr, "Can not find card family\n");
		return 1;
	}

	memset(subchan,0,sizeof(subchan));

	if (rexmlInit(filename_dbxml)<0) {
		fprintf(stderr, "Can not load xml database %s\n", filename_dbxml);
		return 1;
	}

	got_objects = read_objects(filename_objects);

	/* Read subchannel/object from test_startup */
	snprintf(file_startup, sizeof(file_startup), filename_startup,
		(pci_id>>16) & 0xffff, pci_id & 0xffff
	);

	parse_dump(file_startup, 1);

	/* Print list of subchan hash read from test startup */
	printf("/* # Objects hash in subchan after test_startup */\n");
	for (i=0; i<8; i++) {
		if (!subchan[i].name) {
			snprintf(subchan[i].name, SUBCHAN_NAME_SZ,
				"NvObject%04x",
				(int)(get_object_class(i) & 0x7fffffffUL));
		}
		printf("/* #  Subchan[%d].hash = 0x%08x: %s */\n", i, subchan[i].hash, subchan[i].name);
	}

	{
		int new_hash = 0;
		for(i=0; i<argc; i++) {
			if (strcmp("-subchan", argv[i])==0) {
				int n_hash;
				sscanf(argv[i+2], "%x", &n_hash);
				subchan[atoi(argv[i+1])].hash = n_hash << 8;
				new_hash = 1;
			}
		}
		if (new_hash) {
			printf("/* # Objects hash in subchan after changes */\n");
			for (i=0; i<8; i++) {
				if (!subchan[i].name) {
					snprintf(subchan[i].name,
						SUBCHAN_NAME_SZ,
						"NvObject%04x",
						(int)(get_object_class(i) &
							0x7fffffffUL));
				}
				printf("/* #  Subchan[%d].hash = 0x%08x: %s */\n", i, subchan[i].hash, subchan[i].name);
			}
		}
	}

	if (filename_dump) {
		parse_dump(filename_dump, 0);
	}

	if (obj_infos) {
		free(obj_infos);
	}

	rexmlShutdown();
	return 0;
}

void get_pci_id(const char *filename)
{
	FILE *f;

	/* Read pci_id from card_stdout */
	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return;
	}	

	while(!feof(f)) {
		/* Skip comments */
		if (fgetc(f)=='#') {
			int c = 0;
			while(c!='\n') {
				c = fgetc(f);
			}
		}

		int retvalue = fscanf(f, "pci_id\t0x%08x\n", &pci_id);
		if (retvalue == 1) {
			break;
		}
	}
	fclose(f);
}

void get_boot0(const char *filename)
{
	FILE *f;

	/* Read pci_id from card_stdout */
	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return;
	}	

	while(!feof(f)) {
		/* Skip comments */
		if (fgetc(f)=='#') {
			int c = 0;
			while(c!='\n') {
				c = fgetc(f);
			}
		}

		int retvalue = fscanf(f, "BOOT0\t0x%08x\n", &boot0);
		if (retvalue == 1) {
			break;
		}
	}
	fclose(f);
}

int read_objects(const char *filename)
{
	FILE *f;

	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return 0;
	}

	while (!feof(f)) {
		int name, context, data[4];

		fscanf(f, "%08x %08x %08x %08x %08x %08x\n",
			&name, &context, &data[0], &data[1], &data[2], &data[3]);

		add_object(name, context, data);
	}
	fclose(f);

	return 1;
}

void add_object(int name, int context, int *data)
{
	int i = num_obj++;
	obj_infos = realloc(obj_infos, sizeof(obj_info_t) * num_obj);

	obj_infos[i].name = name;
	obj_infos[i].context = context;
	obj_infos[i].data[0] = data[0];
	obj_infos[i].data[1] = data[1];
	obj_infos[i].data[2] = data[2];
	obj_infos[i].data[3] = data[3];

	switch (card_family)
	{
		case NV03:
			obj_infos[i].engine = (obj_infos[i].context >> 23) & 1;
			break;
		case NV04:
		case NV10:
		case NV15:
		case NV20:
		case NV30:
			obj_infos[i].engine = (obj_infos[i].context >> 16) & 0x3;
			break;
		default:
			obj_infos[i].engine = (obj_infos[i].context >> 20) & 0x3;
			break;
	}
}

void printStringDummy(const char *value)
{
}

void printHeaderDummy(int size, int channel, int offset, int increment, int value)
{
}

void printParamDummy(int channel, int offset, int value)
{
	if (offset==0) {
		int object_id;

		subchan[channel].hash = value;
		object_id = get_object_class(channel);
		subchan[channel].node = rexmlGetObjectNode(object_id & 0x7fffffff);
		if (subchan[channel].node) {
			xmlChar *objname = xmlGetProp(
				subchan[channel].node,
				BAD_CAST "name");
			strncpy(subchan[channel].name,
				(char*)objname,
				SUBCHAN_NAME_SZ);
			subchan[channel].name[SUBCHAN_NAME_SZ-1] = '\0';
			xmlFree(objname);
		} else {
			snprintf(subchan[channel].name,
				SUBCHAN_NAME_SZ,
				"NvObject%04x",
				object_id & 0x7fffffff);
		}
	}
}

void printStringReplay(const char *value)
{
	printf("/* %s */\n", value);
}

void printHeaderReplay(int size, int channel, int offset, int increment, int value)
{
	if (offset!=0) {
		int is_array = 0, stride=4;
		int start_offset = offset;

		xmlNodePtr func_node = NULL;
		xmlChar *func_name = NULL;
		xmlChar *parent_name = xmlStrdup(subchan[channel].name);

		/* Object name on this subchannel */
		/*printf("%s", subchan[channel].name);*/

		func_node = rexmlGetFunctionNode(subchan[channel].node, offset, &is_array, &start_offset, &stride);
		if (func_node) {
			func_name = xmlGetProp(func_node, BAD_CAST "name");
			parent_name = xmlGetProp(func_node->parent, BAD_CAST "name");
		}

		printf("\n\tBEGIN_RING(NvSubChan%d, ", channel);
		if (func_name) {
			/* Function name inside object if known */
			printf("%s_%s", parent_name, func_name);
			if (is_array) {
				int array_index = (offset-start_offset)/stride;
				printf("(%d)", array_index);
			}
			xmlFree(func_name);
		} else {
			/* Or offset if unknown */
			printf("0x%04x", offset);
		}
		printf("%s, %d);\n", increment ? "" : " | NONINC_METHOD", size);

		xmlFree(parent_name);
	}
}

void printParamReplay(int channel, int offset, int value)
{
	printParamDummy(channel, offset, value);

	if (offset==0) {
		int object_id = get_object_class(channel);

		printf("\n\tSetSubchannel(NvSubChan%d, 0x%08x);\n", channel, subchan[channel].hash);
		printf("/* %s mapped to subchannel %d %s */\n", subchan[channel].name,
			channel, (object_id & (1<<31)) ? "(GUESS)" : "");
	} else {
		int is_array = 0, stride=4;
		int start_offset = offset;
		int dump_as_float = 0;
		xmlNodePtr func_node = NULL;

		/* Check if float value */
		func_node = rexmlGetFunctionNode(subchan[channel].node, offset,
			&is_array, &start_offset, &stride);
		if (func_node) {
			xmlChar *attValue;

			attValue = xmlGetProp(func_node, BAD_CAST "type");
			if (attValue) {
				dump_as_float = (xmlStrcmp(BAD_CAST "float", attValue)==0);
				xmlFree(attValue);
			}
		}

		if (dump_as_float) {
			float *float_value = (float *) &value;
			printf("\tOUT_RINGf(%f);\n", *float_value);
		} else {
			printf("\tOUT_RING(0x%08x);\n", value);
		}
	}
}

void parse_dump(const char *filename, int readsubchan)
{
	rexmlParseFuncs_t parseFuncs;

	if (readsubchan) {
		parseFuncs.printString = printStringDummy;
		parseFuncs.printHeader = printHeaderDummy;
		parseFuncs.printParam = printParamDummy;
	} else {
		parseFuncs.printString = printStringReplay;
		parseFuncs.printHeader = printHeaderReplay;
		parseFuncs.printParam = printParamReplay;
	}

	rexmlParse(filename, &parseFuncs);
}

int get_object_class(int channel)
{
	if (obj_from_hash) {
		return ((subchan[channel].hash >> 8) & 0xff) | 0x80000000UL;
	}

	return get_object_class_from_hash(subchan[channel].hash);
}

int get_object_class_from_hash(int hash)
{
	/* Search this hash in object list */
	int i, ctx, datas[4] = {0,0,0,0};

	for (i=0; i<num_obj; i++) {
		if (obj_infos[i].name != hash) {
			continue;
		}

		int class_mask = ((1<<12)-1);

		if (obj_infos[i].engine == 1) {
			/* graphics, nv40/g80 has 16bits */
			switch(card_family) {
				case NV40:
				case G80:
					class_mask = ((1<<16)-1);
					break;
				default:
					break;
			}
		}

		return obj_infos[i].data[0] & class_mask;
	}

	/* Create a dummy object from hash */
	datas[0] = (hash>>8) & 0xff;
	ctx = 0x82000000 + datas[0];
	add_object(hash, ctx, datas);

	/* Return object from hash if not found */
	return ((hash >> 8) & 0xff) | 0x80000000UL;
}
